#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "basic.h"

void *malloc_check(size_t sz) {
  if (sz == 0) return NULL ;
  void *p = malloc(sz) ;
  if (!p) {
    perror("malloc") ;
    exit(2) ;
  }
  return p ;
}
/* Threads */
static void exit_error (char *msg, int st) {
  fprintf(stderr,"%s: %s\n",msg,strerror(st));
  exit(EXIT_FAILURE);
}

void create_thread(pthread_t *th,void *(*f)(void *),void *x) {
  int st = pthread_create(th,NULL,f,x);
  if (st != 0) exit_error("pthread_create",st) ;
}

/*
void create_thread_detached(pthread_t *th,void *(*f)(void *),void *x) {
  pthread_attr_t tattr;
  int st ;

  st = pthread_attr_init(&tattr) ;
  if (st != 0) exit_error("pthread_attr_init",st) ;

  st = pthread_attr_setdetachstate(&tattr,PTHREAD_CREATE_DETACHED) ;
  if (st != 0) exit_error("pthread_attr_setdetachstate",st) ;

  st = pthread_create(th,&tattr,f,x) ;
  if (st != 0) exit_error("pthread_create",st) ;

  st = pthread_attr_destroy(&tattr) ;
  if (st != 0) exit_error("pthread_attr_destroy",st) ;
}
*/
void create_thread_detached(pthread_t *th,void *(*f)(void *),void *x) {
  int st ;

  st = pthread_create(th,NULL,f,x) ;
  if (st != 0) exit_error("pthread_create",st) ;
  st = pthread_detach(*th);
  if (st != 0) exit_error("pthread_detach",st) ;
}

void *join_thread(pthread_t *th) {
  void *r ;
  int st = pthread_join (*th,&r) ;
  if (st != 0) exit_error("pthread_join",st) ;
  return r ;
}
/* Mutex */

pthread_mutex_t *alloc_mutex(void) {
  pthread_mutex_t *r = malloc_check(sizeof(*r)) ;
  if (pthread_mutex_init(r,NULL)) {
    perror("pthread_mutex_init") ;
    exit(2) ;
  }
  return r ;
}

void free_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_destroy(p)) {
    perror("pthread_mutex_destroy") ;
    exit(2) ;
  }
  free(p) ;
}

void lock_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_lock(p)) {
    perror("pthread_mutex_lock") ;
    exit(2) ;
  }
}

void unlock_mutex(pthread_mutex_t *p) {
  if (pthread_mutex_unlock(p)) {
    perror("pthread_mutex_unlock") ;
    exit(2) ;
  }
}

/* Condition variable */

pthread_cond_t *alloc_cond(void) {
  pthread_cond_t *r = malloc_check(sizeof(*r)) ;
  if (pthread_cond_init(r,NULL)) {
    perror("pthread_cond_init") ;
    exit(2) ;
   }
  return r ;
}

void free_cond(pthread_cond_t *p) {
  if (pthread_cond_destroy(p)) {
    perror("pthread_cond_destroy") ;
    exit(2) ;
  }
  free(p) ;
}

void wait_cond(pthread_cond_t *c, pthread_mutex_t *m) {
  if (pthread_cond_wait(c,m)) {
    perror("pthread_cond_wait") ;
    exit(2) ;
  }
}

void signal_cond(pthread_cond_t *c) {
  if (pthread_cond_signal(c)) {
    perror("pthread_cond_signal") ;
    exit(2) ;
  }
}
void broadcast_cond(pthread_cond_t *c) {
  if (pthread_cond_broadcast(c)) {
    perror("pthread_cond_broadcast") ;
    exit(2) ;
  }
}
