#include <stdio.h>
#include <stdlib.h>
#include "basic.h"

typedef struct {  int v ; } val_t ;

val_t *alloc_val(int i) {
  val_t *r = malloc_check(sizeof(*r)) ;
  r->v = i ;
  return r ;
}

void free_val(val_t *p) {
  free(p) ;
}

int square(int i) { return i*i ; }

void *f(void *_p) {
  val_t *p = (val_t *)_p ;
  int i = p->v ;
  free_val(p) ;
  int j = square(i) ;
  return alloc_val(j) ;
}

int sum(int n) {
  // Fork
  pthread_t th[n] ;
  for (int k = 0 ; k < n ;  k++)
    create_thread(&th[k],f,alloc_val(k+1)) ;

  // Retrieve and sum results
  int r = 0 ;
  for (int k = 0 ; k < n ;  k++) {
    val_t *p = (val_t *)join_thread(&th[k]) ;
    r += p->v ;
    free_val(p) ;
  }
  return r ;
}

/* Update share memory */

static volatile int sq ;

void *f2(void *p) {
  int i = ((val_t *)p)->v ;
  sq += i*i ;
  return NULL ;
}

int sum2(int n) {
  sq = 0 ; // Be cautious
  // Fork
  pthread_t th[n] ;
  val_t arg[n] ;
  for (int k = 0 ; k < n ;  k++) {
    val_t *a = &arg[k] ;
    a->v = k+1 ;
    create_thread(&th[k],f2,a) ;
  }

  // Wait all threads
  for (int k = 0 ; k < n ;  k++)
    (void)join_thread(&th[k]) ;

  return sq ;
}

/* With lock */

pthread_mutex_t *mutex ;

void *f3(void *p) {
  int i = ((val_t *)p)->v ;
  int j = i*i ;
  lock_mutex(mutex) ;
  sq += j ;
  unlock_mutex(mutex) ;
  return NULL ;
}

int sum3(int n) {
  sq = 0 ; // Be cautious
  mutex = alloc_mutex() ;
  // Fork
  pthread_t th[n] ;
  val_t arg[n] ;
  for (int k = 0 ; k < n ;  k++) {
    val_t *a = &arg[k] ;
    a->v = k+1 ;
    create_thread(&th[k],f3,a) ;
  }

  // Wait all threads
  for (int k = 0 ; k < n ;  k++)
    (void)join_thread(&th[k]) ;

  free_mutex(mutex) ;
  return sq ;
}

int main(int argc, char **argv) {
  int n = 10 ;
  if (argc > 1) n = atoi(argv[1]) ;

  int r1 = sum(n) ;
  printf("R1: %i\n",r1) ;
  int r2 = sum2(n) ;
  printf("R2: %i\n",r2) ;
  int r3 = sum3(n) ;
  printf("R3: %i\n",r3) ;

  return 0 ;
}
