#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "basic.h"
#include "queen.h"

/*****************/
/* Boxed count_t */
/*****************/

typedef struct {
  count_t c ;
} val_t ;

void free_val(val_t *p) { free(p) ; }

val_t *alloc_val(count_t c) {
  val_t *r = malloc_check(sizeof(*r)) ;
  r->c = c ;
  return r ;
}


/**********************************/
/* Fork a thread to run a subtask */
/**********************************/

void *run_stub(void *z) {
  count_t r = run_subtask((subtask_t *)z) ;
  free_arg(z) ;
// By argreement, this was my own copy (See emit_thread below)
  return alloc_val(r) ;
}

// NB: allocates thread and returns it
pthread_t *fork_subtask(subtask_t *z) {
  pthread_t *th = malloc_check(sizeof(*th)) ;
  create_thread(th,run_stub,z) ;
  return th ;
}

/*******************/
/* List of threads */
/*******************/

typedef struct cell {
  pthread_t *fst ;
  struct cell *next ;
} cell_t ;

void free_cell(cell_t *p) { return free(p) ; }

cell_t *alloc_cell(pthread_t *fst, cell_t *next) {
  cell_t *r = malloc(sizeof(*r)) ;
  if (!r) {
    perror("malloc") ;
    exit(2) ;
  }
  r->fst = fst ;
  r->next = next ;
  return r ;
}


/* Lisy of create threads */
static cell_t *threads = NULL ;

void emit_thread(subtask_t *z) {
  z = copy_arg(z) ; // Be cautious..
  threads = alloc_cell(fork_subtask(z),threads) ;  
}

count_t run(int n, int depth) {
  count_t c = 0 ;

  /* Fork all subtasks */
  int ntasks = generate_subtasks(n,depth,emit_thread) ;
  if (verbose) fprintf(stderr,"Subtasks: %i\n",ntasks) ;

  /* Sum results */
  cell_t *q = NULL ;
  for (cell_t *p = threads ; p ;
       q = p->next, free_cell(p), p = q) {
    void *r = join_thread(p->fst) ;
    free(p->fst) ;
    val_t *_r = (val_t *)r ;
    c += _r->c ;
    free_val(_r) ;
  }
  return c ;
}

void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v|-d<int>]* [n]\n",pgm);
  exit(-1);
}

int main(int argc, char **argv) {
  char *arg=0;
  char *pgm = argv[0];
  int n = 8 ;
  int depth = 3 ;

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
       if (*arg == 'd') {
        if (arg[1]) {
          depth = atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
       } else if (*arg == 'v') {
         verbose++;
       } else {
         usage(pgm);
       }
    } else break ;      
  }
  if (argc >= 1) n = atoi(arg);
  if (n > NMAX) {
      fprintf(stderr,"argument must be <= %zi\n",NMAX);
      exit(2);
  }
  count_t r = run(n,depth) ;
  printf("%" PRC "\n",2*r) ;
  return 0 ;
}
